--[[-------------------------------------------------------------------------
	InterruptHelper
	Copyright (C) 2011  Morsker
	All Rights Reserved.
	Please contact Morsker through PM on wowace.com.

	Helps with interrupt notifications and announcements.
---------------------------------------------------------------------------]]

local InterruptHelper = InterruptHelper
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale("InterruptHelper")
local LSM = LibStub("LibSharedMedia-3.0")
local db = InterruptHelper.db
local options

-- GLOBALS: LibStub
-- GLOBALS: INTERFACEOPTIONS_ADDONCATEGORIES
-- GLOBALS: InterfaceAddOnsList_Update
-- GLOBALS: tremove
local UnitExists = UnitExists
local UnitIsPlayer = UnitIsPlayer
local UnitName = UnitName
local floor = floor
local format = format
local pairs = pairs
local sort = sort
local strmatch = strmatch
local wipe = wipe

local annSayFormatDesc = [[
Format for announcements.
$s = spell name
$S = spell link
]]

local raidChatValues = {
	["NONE"] = "None",
	["TEXT"] = "Text",
	["SAY"] = "Say",
	["WHISPER"] = "Whisper",
	["CHANNEL"] = "Channel",
	["PARTY"] = "Party",
	["RAID"] = "Raid",
}

local partyChatValues = {
	["NONE"] = "None",
	["TEXT"] = "Text",
	["SAY"] = "Say",
	["WHISPER"] = "Whisper",
	["CHANNEL"] = "Channel",
	["PARTY"] = "Party",
}

local soloChatValues = {
	["NONE"] = "None",
	["TEXT"] = "Text",
	["SAY"] = "Say",
	["WHISPER"] = "Whisper",
	["CHANNEL"] = "Channel",
}

local battlegroundChatValues = {
	["NONE"] = "None",
	["TEXT"] = "Text",
	["SAY"] = "Say",
	["WHISPER"] = "Whisper",
	["CHANNEL"] = "Channel",
	["BATTLEGROUND"] = "Battleground",
}

local function sortedKeys(hashTable, out)
	if not out then out = {} end
	for k,_ in pairs(hashTable) do
		out[#out+1] = k
	end
	sort(out)
	return out
end

local function SetGeneric(info, val)
	local setting = info[#info]
	db.profile[setting] = val
end

local function ValidateGeneric(info, val)
	local invalid = strmatch(val, "[,_|]")
	return invalid and format("Invalid character: %s", invalid) or true
end

local function DisabledChannel(info)
	return not db.profile.annAnnounce
end

local function MakeGap(order)
	return {
		order = order,
		type = "description",
		name = "",
		width = "full",
	}
end

local function MakeChatSelect(order, name, chatValues)
	return {
		type = "select",
		name = name,
		order = order,
		values = chatValues,
		disabled = DisabledChannel,
		set = SetGeneric,
	}
end

local function MakeChatArg(order, base)
	return {
		type = "input",
		name = function(info)
			if db.profile[base] == "WHISPER" then
				return "Player"
			elseif db.profile[base] == "CHANNEL" then
				return "Channel"
			end
		end,
		order = order,
		disabled = DisabledChannel,
		hidden = function(info)
			return not (db.profile[base] == "WHISPER" or db.profile[base] == "CHANNEL")
		end,
		get = function(info)
			if db.profile[base] == "WHISPER" then
				return db.profile[base.."Player"]
			elseif db.profile[base] == "CHANNEL" then
				return db.profile[base.."Channel"]
			end
		end,
		set = function(info, val)
			if db.profile[base] == "WHISPER" then
				db.profile[base.."Player"] = val
			elseif db.profile[base] == "CHANNEL" then
				db.profile[base.."Channel"] = val
			end
		end,
	}
end

local function GetTracking(info)
	local mobname = info[#info-1]
	local spell = info[#info]
	return db.profile.tracking[mobname][spell]
end

local function SetTracking(info, val)
	local mobname = info[#info-1]
	local spell = info[#info]
	db.profile.tracking[mobname][spell] = val
end

local function GetTrackAll(info)
	local mobname = info[#info-1]
	local t = db.profile.tracking[mobname]
	for k,v in pairs(t) do
		if (not strmatch(k, "^_")) and not v then
			return false
		end
	end
	return true
end

local function HideTrackAll(info)
	local mobname = info[#info-1]
	for k,_ in pairs(db.profile.tracking[mobname]) do
		if not strmatch(k, "^_") then
			return false
		end
	end
	return true
end

local function GetTrackNone(info)
	local mobname = info[#info-1]
	local t = db.profile.tracking[mobname]
	for k,v in pairs(t) do
		if (not strmatch(k, "^_")) and v then
			return false
		end
	end
	return true
end

local function SetTrackAll(info)
	local mobname = info[#info-1]
	local t = db.profile.tracking[mobname]
	for k,_ in pairs(t) do
		if not strmatch(k, "^_") then
			t[k] = true
		end
	end
end

local function SetTrackNone(info)
	local mobname = info[#info-1]
	local t = db.profile.tracking[mobname]
	for k,_ in pairs(t) do
		if not strmatch(k, "^_") then
			t[k] = false
		end
	end
end

local function DescTrackingSpellView(info)
	local spell = info[#info]
	return format("Warn when %s is cast.", spell)
end

function InterruptHelper:NewTrackingSpellView(name, spell)
	options.args.tracking.args[name].args[spell] = {
		order = 10, -- FIXME
		type = "toggle",
		name = spell,
		desc = DescTrackingSpellView,
	}
end

local _track_head1 = {
	order = 1,
	type = "header",
	name = function(info)
		return info[#info-1]
	end,
}

local _trackAddSpell = {
	order = 2,
	type = "input",
	name = "Add Spell (case-sensitive):",
	desc = function(info)
		local mobname = info[#info-1]
		return format("Add a spell %1$s can cast. The name must match exactly.", mobname)
	end,
	validate = ValidateGeneric,
	get = false,
	set = function(info, val)
		if val and (#val > 0) then
			local mobname = info[#info-1]
			db.profile.tracking[mobname][val] = true
			InterruptHelper:NewTrackingSpellView(mobname, val)
			InterruptHelper:OrderTrackingView()
		end
	end,
}

local _trackRemoveSpell = {
	order = 3,
	type = "select",
	name = "Remove Spell:",
	desc = "Choose a spell to remove. This is to fix mistakes if you accidentally add the wrong spell name. If you want to disable warnings for a spell, you can uncheck it instead of removing it.",
	hidden = HideTrackAll,
	values = function(info)
		local mobname = info[#info-1]
		local t = {}
		for k,_ in pairs(db.profile.tracking[mobname]) do
			if not strmatch(k, "^_") then
				t[k] = k
			end
		end
		return t
	end,
	get = false,
	set = function(info, val)
		local mobname = info[#info-1]
		db.profile.tracking[mobname][val] = nil
		options.args.tracking.args[mobname].args[val] = nil
	end,
}

local _track_gap1 = MakeGap(4)

local _trackLearn = {
	order = 5,
	type = "toggle",
	name = "Learn Unknown",
	desc = function(info)
		local mobname = info[#info-1]
		return format([[
Watch and learn names of unknown spells %1$s casts.

Gold: %2$s
Empty: %3$s
Gray: Default (%4$s)]], mobname, "Yes", "No", db.profile.trackLearn and "Yes" or "No")
	end,
	get = function(info)
		local mobname = info[#info-1]
		return db.profile.tracking[mobname]._learn
	end,
	set = function(info, val)
		local mobname = info[#info-1]
		db.profile.tracking[mobname]._learn = val
	end,
	tristate = true,
}

local _trackWarnUnknown = {
	order = 6,
	type = "toggle",
	name = "Warn For Unknown",
	desc = function(info)
		local mobname = info[#info-1]
		return format([[
Warn for all spells %1$s casts, unless the spell is known and unchecked.

Gold: %2$s
Empty: %3$s
Gray: Default (%4$s)]], mobname, "Yes", "No", db.profile.trackWarnUnknown and "Yes" or "No")
	end,
	get = function(info)
		local mobname = info[#info-1]
		return db.profile.tracking[mobname]._warnUnknown
	end,
	set = function(info, val)
		local mobname = info[#info-1]
		db.profile.tracking[mobname]._warnUnknown = val
	end,
	tristate = true,
}

local _track_head2 = {
	order = 7,
	type = "header",
	name = "",
	hidden = HideTrackAll,
}

local _trackAll = {
	order = 8,
	type = "toggle",
	name = "All",
	desc = "Select all.",
	hidden = HideTrackAll,
	get = GetTrackAll,
	set = SetTrackAll,
}

local _trackNone = {
	order = 9,
	type = "toggle",
	name = "None",
	desc = "Unselect all.",
	hidden = HideTrackAll,
	get = GetTrackNone,
	set = SetTrackNone,
}

local _track_head3 = {
	order = 998,
	type = "header",
	name = "",
}

local _trackRemove = {
	order = 999,
	type = "execute",
	name = "Remove",
	desc = function(info)
		local mobname = info[#info-1]
		return format("Remove %1$s and all spells and settings for %1$s.", mobname)
	end,
	func = function(info)
		local mobname = info[#info-1]
		db.profile.tracking[mobname] = nil
		options.args.tracking.args[mobname] = nil
	end,
}

function InterruptHelper:OrderTrackingView()
	local mobnames = options.args.tracking.args
	local keys = sortedKeys(mobnames)
	local keys2
	local order = 100
	for i = 1,#keys do
		local k = keys[i]
		if not strmatch(k, "^_") then
			mobnames[k].order = order
			order = order + 1
			local spellnames = mobnames[k].args
			keys2 = keys2 and wipe(keys2) or {}
			keys2 = sortedKeys(spellnames, keys2)
			local order2 = 100
			for i2 = 1,#keys2 do
				local k2 = keys2[i2]
				if not strmatch(k2, "^_") then
					spellnames[k2].order = order2
					order2 = order2 + 1
				end
			end
		end
	end
end

local function NewTracking(name)
	if db.profile.tracking[name] then return end
	db.profile.tracking[name] = {}
end

local function NewTrackingView(name)
	if options.args.tracking.args[name] then return end
	local settings = db.profile.tracking[name]
	options.args.tracking.args[name] = {
		order = 10, -- FIXME
		type = "group",
		name = name,
		args = {
			_track_head1 = _track_head1,
			_trackAddSpell = _trackAddSpell,
			_trackRemoveSpell = _trackRemoveSpell,
			_track_gap1 = _track_gap1,
			_trackLearn = _trackLearn,
			_trackWarnUnknown = _trackWarnUnknown,
			_track_head2 = _track_head2,
			_trackAll = _trackAll,
			_trackNone = _trackNone,
			-- _track_gap1 = _track_gap1,
			-- spells end up here
			_track_head3 = _track_head3,
			_trackRemove = _trackRemove,
		},
	}
	for spell,_ in pairs(settings) do
		if not strmatch(spell, "^_") then
			InterruptHelper:NewTrackingSpellView(name, spell)
		end
	end
end

function InterruptHelper:InitOptions()
	options = {
		type = "group",
		name = "Interrupt Helper",
		childGroups = "tab",
		get = function(info)
			return db.profile[info[#info]]
		end,
		args = {
			tracking = {
				order = 10,
				type = "group",
				name = "Tracking",
				childGroups = "tree",
				get = GetTracking,
				set = SetTracking,
				args = {
					_add = {
						order = 1,
						type = "input",
						name = "Track Target (case-sensitive):",
						desc = "Enter the name of a target to track, so InterruptHelper will show warnings for it. The name must match exactly.",
						width = "double",
						validate = ValidateGeneric,
						get = false,
						set = function(info, val)
							if val and (#val > 0) then
								NewTracking(val)
								NewTrackingView(val)
								InterruptHelper:OrderTrackingView()
								LibStub("AceConfigDialog-3.0"):SelectGroup("InterruptHelper", "tracking", val)
							end
						end,
					},
					_addTarget = {
						order = 2,
						type = "execute",
						name = "Use Target",
						desc = "Use the name of your current target.",
						func = function(info)
							if UnitExists("target") then
								local name = UnitName("target")
								NewTracking(name)
								NewTrackingView(name)
								InterruptHelper:OrderTrackingView()
								LibStub("AceConfigDialog-3.0"):SelectGroup("InterruptHelper", "tracking", name)
							end
						end,
					},
					_gap1 = MakeGap(3),
					_trackLearn = {
						order = 4,
						type = "toggle",
						name = "Learn Unknown",
						desc = format([[
When tracking a target, watch and learn names of unknown spells.

%1$s]], "(This setting is the default for all targets. It can be overridden on a per-target basis.)"),
						get = function(info)
							return db.profile.trackLearn
						end,
						set = function(info, val)
							db.profile.trackLearn = val
						end,
					},
					_trackWarnUnknown = {
						order = 5,
						type = "toggle",
						name = "Warn For Unknown",
						desc = format([[
Warn for all spells cast by tracked targets, unless the spell is known and unchecked.

%1$s]], "(This setting is the default for all targets. It can be overridden on a per-target basis.)"),
						get = function(info)
							return db.profile.trackWarnUnknown
						end,
						set = function(info, val)
							db.profile.trackWarnUnknown = val
						end,
					},
					_trackWarnAll = {
						order = 6,
						type = "toggle",
						name = "Warn For All Targets",
						desc = "Warn for all targets, even when not tracked. This can be annoying, but is sometimes useful for testing.",
						get = function(info)
							return db.profile.trackWarnAll
						end,
						set = function(info, val)
							db.profile.trackWarnAll = val
						end,
					},
				},
			},
			warnings = {
				order = 20,
				type = "group",
				name = "Warnings",
				args = {
					warnSimultaneous = {
						order = 1,
						type = "select",
						name = "If Target = Focus",
						desc = "If your focus and target are the same, don't play both warnings at once; only play one. Which one?",
						values = { ["target"] = "Target", ["focus"] = "Focus", },
						set = SetGeneric,
					},
					headerTarget = {
						order = 100,
						type = "header",
						name = "Target",
					},
					targetEnabled = {
						order = 110,
						type = "toggle",
						name = "Enabled",
						set = SetGeneric,
					},
					testTg = {
						order = 120,
						type = "execute",
						name = "Test",
						func = function(info)
							InterruptHelper:Warn("target")
						end,
					},
					targetFlashRGB = {
						order = 130,
						type = "color",
						name = "Color",
						hasAlpha = false,
						get = function(info)
							local prof = db.profile
							return prof.targetFlashR, prof.targetFlashG, prof.targetFlashB
						end,
						set = function(info, r, g, b)
							local prof = db.profile
							prof.targetFlashR = r
							prof.targetFlashG = g
							prof.targetFlashB = b
						end,
					},
					targetFlashA = {
						order = 140,
						type = "range",
						name = "Alpha",
						min = 0,
						max = 1,
						softMin = 0.2,
						softMax = 0.5,
						step = 0.01,
						bigStep = 0.05,
						set = function(info, val)
							db.profile.targetFlashA = val
						end,
					},
					targetFlashTime = {
						order = 150,
						name = "Duration",
						type = "range",
						min = 0,
						max = 2,
						softMin = 0.0,
						softMax = 0.5,
						step = 0.01,
						bigStep = 0.05,
						set = function(info, val)
							db.profile.targetFlashTime = val
						end,
					},
					targetSound = {
						order = 160,
						type = "select",
						dialogControl = "LSM30_Sound",
						name = "Sound",
						values = InterruptHelper.definedSounds,
						get = function(info)
							return db.profile.targetSound
						end,
						set = function(info, val)
							db.profile.targetSound = val
						end,
					},
					headerFocus = {
						order = 200,
						type = "header",
						name = "Focus",
					},
					focusEnabled = {
						order = 210,
						type = "toggle",
						name = "Enabled",
						set = SetGeneric,
					},
					testFo = {
						order = 220,
						type = "execute",
						name = "Test",
						func = function(info)
							InterruptHelper:Warn("focus")
						end,
					},
					focusFlashRGB = {
						order = 230,
						type = "color",
						name = "Color",
						hasAlpha = false,
						get = function(info)
							local prof = db.profile
							return prof.focusFlashR, prof.focusFlashG, prof.focusFlashB
						end,
						set = function(info, r, g, b)
							local prof = db.profile
							prof.focusFlashR = r
							prof.focusFlashG = g
							prof.focusFlashB = b
						end,
					},
					focusFlashA = {
						order = 240,
						type = "range",
						name = "Alpha",
						min = 0,
						max = 1,
						softMin = 0.2,
						softMax = 0.5,
						step = 0.01,
						bigStep = 0.05,
						set = function(info, val)
							db.profile.focusFlashA = val
						end,
					},
					focusFlashTime = {
						order = 250,
						name = "Duration",
						type = "range",
						min = 0,
						max = 2,
						softMin = 0.0,
						softMax = 0.5,
						step = 0.01,
						bigStep = 0.05,
						set = function(info, val)
							db.profile.focusFlashTime = val
						end,
					},
					focusSound = {
						order = 260,
						type = "select",
						dialogControl = "LSM30_Sound",
						name = "Sound",
						values = InterruptHelper.definedSounds,
						get = function(info)
							return db.profile.focusSound
						end,
						set = function(info, val)
							db.profile.focusSound = val
						end,
					},
				},
			},
			announce = {
				order = 30,
				type = "group",
				name = "Announce Settings",
				args = {
					annAnnounce = {
						type = "toggle",
						name = "Announce Interrupts",
						desc = "Announce interrupts in the channels of your choice.",
						order = 10,
						set = function(info, val)
							SetGeneric(info, val)
							if val then
								InterruptHelper.cleuFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
							else
								InterruptHelper.cleuFrame:UnregisterAllEvents()
							end
						end
					},
					annSayFormat = {
						type = "input",
						name = "Format",
						desc = annSayFormatDesc,
						order = 20,
						width = "double",
						disabled = DisabledChannel,
						set = SetGeneric,
					},
					_gap_1 = MakeGap(100),
					annSolo = MakeChatSelect(110, "Solo", soloChatValues),
					annSoloArg = MakeChatArg(120, "annSolo"),
					_gap_2 = MakeGap(200),
					annParty = MakeChatSelect(210, "Party", partyChatValues),
					annPartyArg = MakeChatArg(220, "annParty"),
					_gap_3 = MakeGap(300),
					annRaid = MakeChatSelect(310, "Raid", raidChatValues),
					annRaidArg = MakeChatArg(320, "annRaid"),
					_gap_4 = MakeGap(400),
					annBattleground = MakeChatSelect(410, "Battleground", battlegroundChatValues),
					annBattlegroundArg = MakeChatArg(420, "annBattleground"),
					_gap_5 = MakeGap(500),
					annCombatZone = MakeChatSelect(510, "Combat Zone", raidChatValues),
					annCombatZoneArg = MakeChatArg(520, "annCombatZone"),
				},
			},
			profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(db),
		},
	}
	InterruptHelper.options = options

	-- populate options/tracking from db.profile.tracking
	for name,_ in pairs(db.profile.tracking) do
		NewTrackingView(name)
	end
	InterruptHelper:OrderTrackingView()

	-- Remove any stub in the interface options, before adding the real entry
	local categories = INTERFACEOPTIONS_ADDONCATEGORIES
	for i = 1,#categories do
		if categories[i].name == "InterruptHelper" then
			tremove(categories, i)
			break
		end
	end
	InterfaceAddOnsList_Update()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("InterruptHelper", options)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("InterruptHelper_" .. options.args.tracking.name, options.args.tracking)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("InterruptHelper_" .. options.args.warnings.name, options.args.warnings)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("InterruptHelper_" .. options.args.announce.name, options.args.announce)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("InterruptHelper_" .. options.args.profile.name, options.args.profile)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("InterruptHelper", "InterruptHelper")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("InterruptHelper_" .. options.args.tracking.name, options.args.tracking.name, "InterruptHelper")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("InterruptHelper_" .. options.args.warnings.name, options.args.warnings.name, "InterruptHelper")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("InterruptHelper_" .. options.args.announce.name, options.args.announce.name, "InterruptHelper")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("InterruptHelper_" .. options.args.profile.name, options.args.profile.name, "InterruptHelper")
end
